// power_branchAlways_LRAndR.C
// See the .h file for comments

#include "power_branch_LRAndR.h"
#include "power_instr.h"
#include "directToMemoryEmitter.h"
#include "funkshun.h"


branchAlways_LRAndR::branchAlways_LRAndR(kptr_t iLaunchSiteAddr,
					 const regset_t *iFreeIntRegsBeforeInsn,
					 const regset_t *iFreeIntRegsAfterInsn,
					 const instr_t *ioriginsn,
					 SpringBoardHeap &isbheap) : 
  LRAndR(iLaunchSiteAddr,
	 iFreeIntRegsBeforeInsn,
	 iFreeIntRegsAfterInsn,
	 ioriginsn, 
	 isbheap)
{
   power_instr::power_cfi cfi;
   ioriginsn->getControlFlowInfo(&cfi);
   assert(cfi.fields.controlFlowInstruction && !cfi.power_fields.condTrue);
   if (cfi.destination == pcRelative)
     branch_destination = iLaunchSiteAddr + cfi.offset_nbytes;
   else {
     assert(cfi.destination == absolute);
     branch_destination = cfi.offset_nbytes;
   }
       
}

unsigned branchAlways_LRAndR::conservative_numInsnBytesForRAndR() const {
  return 4;
  //this will remain a simple jump, no matter what
}

void branchAlways_LRAndR::
registerPatchAddrAndCreateLauncher(kptr_t ipatchAddr,
                                   const instr_t *whenAllDoneRestoreToThisInsn) {
   setPatchAddr(ipatchAddr);
   setLauncherToAnyAnnulledLauncher(patchAddr, whenAllDoneRestoreToThisInsn);
}

void branchAlways_LRAndR::emitStuffAfterPreInsnSnippets(directToMemoryEmitter_t &em,
                                                const function_t &fn,
                                                const instr_t::controlFlowInfo*
                                                ) const {
   const unsigned initialNumInsnBytesEmitted = em.numInsnBytesEmitted();
   kptr_t currInsnAddr = patchAddr + em.numInsnBytesEmitted();
   bool relSuccess = false;
   instr_t *relocatedBranch = 
     fn.get1OrigInsn(launchSiteAddr)->tryRelocationNoLink(launchSiteAddr, currInsnAddr, &relSuccess);
   if (relSuccess)
     em.emit(relocatedBranch);
   else {
     //double check that we really are out of range
     assert(!power_instr::inRangeOfBranchCond(launchSiteAddr, currInsnAddr));
     assert(cfi.destination == pcRelative); //if absolute, how can it be out of range ?
     //this better work
     assert(power_instr::inRangeOfBranch(launchSiteAddr, currInsnAddr));
     
     //replace a "conditional" branch always with a long jump
     delete relocatedBranch; //useless now

     kaddrdiff_t new_displacement = launchSiteAddr + offset - currInsnAddr;
     simmediate<24> li = new_displacement/4; //number of words
     relocatedBranch = new power_instr(power_instr::branch, li, 0/*not absolute */ , 0 /*no linking */); 
     em.emit(relocatedBranch);
   }

   assert(em.numInsnBytesEmitted() - initialNumInsnBytesEmitted <= conservative_numInsnBytesForRAndR());
}

